#ifndef __PLUGIN_RENDERER_CPVR_RENDERERCPVR_H
#define __PLUGIN_RENDERER_CPVR_RENDERERCPVR_H

#include "volumeshop.h"

#include <list>
#include <set>
#include "Environment.h"
#include "Volume.h"
#include "Renderer.h"
#include "Vector.h"
#include "Quaternion.h"
#include "Matrix.h"
#include "Shader.h"
#include "Color.h"
#include "Timer.h"
#include "Label.h"
#include "ColorTransferFunction.h"
#include "LightingTransferFunction.h"
#include "Exception.h"
#include "Font.h"

class RendererCPVR : public Renderer
{
public:

	RendererCPVR(Environment & envEnvironment) : m_envEnvironment(envEnvironment), m_uWidth(32), m_uHeight(32),
		m_uVolumeFrameBuffer(0), m_uVolumeColorTexture(0), m_uVolumeDepthRenderBuffer(0), m_uFrontColorTexture(0), m_uBackColorTexture(0),
		m_proVolume(VertexShader("RendererCPVRVolumeVertex.glsl"),FragmentShader("RendererCPVRVolumeFragment.glsl")),
		m_proEntry(VertexShader("RendererCPVREntryVertex.glsl"),FragmentShader("RendererCPVREntryFragment.glsl"))
	{
		if (!GLEW_ARB_fragment_shader)
		{
			std::cerr << "GL_ARB_fragment_shader is not supported by your hardware or driver." << std::endl;
			throw Exception("GL_ARB_fragment_shader is not supported by your hardware or driver.");
		}

		if (!GLEW_ARB_vertex_shader)
		{
			std::cerr << "GL_ARB_fragment_shader is not supported by your hardware or driver." << std::endl;
			throw Exception("GL_ARB_fragment_shader is not supported by your hardware or driver.");
		}

		if (!GLEW_ARB_texture_non_power_of_two)
		{
			std::cerr << "GL_ARB_texture_non_power_of_two is not supported by your hardware or driver." << std::endl;
			throw Exception("GL_ARB_texture_non_power_of_two is not supported by your hardware or driver.");
		}

		if (!GLEW_ARB_texture_rectangle)
		{
			std::cerr << "GL_ARB_texture_rectangle is not supported by your hardware or driver." << std::endl;
			throw Exception("GL_ARB_texture_rectangle is not supported by your hardware or driver.");
		}		

		if (!GLEW_EXT_framebuffer_object)
		{
			std::cerr << "GL_EXT_framebuffer_object is not supported by your hardware or driver." << std::endl;
			throw Exception("GL_EXT_framebuffer_object is not supported by your hardware or driver.");
		}


		glGenFramebuffersEXT(1, &m_uVolumeFrameBuffer);

		glGenTextures(1,&m_uVolumeColorTexture);
		glGenTextures(1,&m_uFrontColorTexture);
		glGenTextures(1,&m_uBackColorTexture);

		glGenRenderbuffersEXT(1, &m_uVolumeDepthRenderBuffer);  
	};

	virtual ~RendererCPVR()
	{
		if (m_uFrontColorTexture)
			glDeleteTextures(1,&m_uFrontColorTexture);

		if (m_uBackColorTexture)
			glDeleteTextures(1,&m_uBackColorTexture);

		if (m_uVolumeColorTexture)
			glDeleteTextures(1,&m_uVolumeColorTexture);

		if (m_uVolumeDepthRenderBuffer)
			glDeleteRenderbuffersEXT(1, &m_uVolumeDepthRenderBuffer);

		if (m_uVolumeFrameBuffer)
			glDeleteFramebuffersEXT(1, &m_uVolumeFrameBuffer);
	};

	Environment & GetEnvironment()
	{
		return m_envEnvironment;
	};

	virtual void idle()
	{
	};

	virtual void underlay()
	{
	};

	virtual void display()
	{
		glPushAttrib(GL_ALL_ATTRIB_BITS);
	
		glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, m_uVolumeFrameBuffer);

		glClearColor(0.0f,0.0f,0.0f,0.0f);
		glClear(GL_COLOR_BUFFER_BIT);
		renderBackgroundVolume();
		
		glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);

		glEnable(GL_BLEND);
		glBlendFunc(GL_ONE, GL_ONE_MINUS_SRC_ALPHA);

		glMatrixMode(GL_PROJECTION);
		glPushMatrix();
		glLoadIdentity();

		glMatrixMode(GL_MODELVIEW);
		glPushMatrix();
		glLoadIdentity();

		glActiveTexture(GL_TEXTURE0);

		glEnable(GL_TEXTURE_2D);
		glBindTexture(GL_TEXTURE_2D,m_uVolumeColorTexture);

		glBegin(GL_QUADS);

		glTexCoord2i(0.0f,0.0f);
		glVertex2f(-1.0f,-1.0f);

		glTexCoord2f(1.0f,0.0f);
		glVertex2f(1.0f,-1.0f);

		glTexCoord2f(1.0f,1.0f);
		glVertex2f(1.0f,1.0f);

		glTexCoord2f(0.0f,1.0f);
		glVertex2f(-1.0f,1.0f);

		glEnd();

		glBindTexture(GL_TEXTURE_2D,0);
	
		glMatrixMode(GL_PROJECTION);
		glPopMatrix();

		glMatrixMode(GL_MODELVIEW);
		glPopMatrix();

		// FPS counter
		static Timer timTimer;
		const float fTime = timTimer;
		timTimer.start();
		static Font fonFont("Arial",10);
		glColor4f(0.0f,0.0f,0.5f,1.0f);
		glRasterPos2f(-0.995f,-0.995f);
		fonFont.display("NEW: %03d",int(1.0f/fTime));
		//

		glPopAttrib();
	};

	virtual void overlay()
	{
	};

	virtual void reshape(const unsigned int uWidth, const unsigned int uHeight)
	{
		GLenum glStatus;

		m_uWidth = uWidth;
		m_uHeight = uHeight;

		//////////////////////////////////////////////////////////////////////////
		glBindTexture(GL_TEXTURE_2D,m_uVolumeColorTexture);
		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_CLAMP);
		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_CLAMP);
		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
		glTexImage2D(GL_TEXTURE_2D,0,GL_RGBA,uWidth,uHeight,0,GL_RGBA,GL_UNSIGNED_BYTE,NULL);
		glBindTexture(GL_TEXTURE_2D,0);
		//////////////////////////////////////////////////////////////////////////
		glBindTexture(GL_TEXTURE_2D,m_uFrontColorTexture);
		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_CLAMP);
		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_CLAMP);
		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
		glTexImage2D(GL_TEXTURE_2D,0,GL_RGBA,uWidth,uHeight,0,GL_RGBA,GL_UNSIGNED_BYTE,NULL);
		glBindTexture(GL_TEXTURE_2D,0);
		//////////////////////////////////////////////////////////////////////////
		glBindTexture(GL_TEXTURE_2D,m_uBackColorTexture);
		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_CLAMP);
		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_CLAMP);
		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
		glTexImage2D(GL_TEXTURE_2D,0,GL_RGBA,uWidth,uHeight,0,GL_RGBA,GL_UNSIGNED_BYTE,NULL);
		glBindTexture(GL_TEXTURE_2D,0);
		//////////////////////////////////////////////////////////////////////////
		glBindRenderbufferEXT(GL_RENDERBUFFER_EXT, m_uVolumeDepthRenderBuffer);
		glRenderbufferStorageEXT(GL_RENDERBUFFER_EXT,GL_DEPTH_COMPONENT, uWidth, uHeight);  
		glBindRenderbufferEXT(GL_RENDERBUFFER_EXT, 0);
		//////////////////////////////////////////////////////////////////////////
		glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, m_uVolumeFrameBuffer);
		glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT,GL_COLOR_ATTACHMENT0_EXT,GL_TEXTURE_2D, m_uVolumeColorTexture, 0);
		glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT,GL_COLOR_ATTACHMENT1_EXT,GL_TEXTURE_2D, m_uFrontColorTexture, 0);
		glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT,GL_COLOR_ATTACHMENT2_EXT,GL_TEXTURE_2D, m_uBackColorTexture, 0);
		glFramebufferRenderbufferEXT(GL_FRAMEBUFFER_EXT,GL_DEPTH_ATTACHMENT_EXT,GL_RENDERBUFFER_EXT, m_uVolumeDepthRenderBuffer);  		

		glStatus = glCheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT);
		if (glStatus  != GL_FRAMEBUFFER_COMPLETE_EXT)
		{
			std::cerr << "Error initializing volume framebuffer." << std::endl << std::endl;
			throw Exception("Error initializing volume framebuffer.");
		}

		glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);
		//////////////////////////////////////////////////////////////////////////
	};


	virtual void mousePress(const MouseEvent & mouEvent)
	{
	};

	virtual void mouseRelease(const MouseEvent & mouEvent)
	{	
	};

	virtual void mouseMove(const MouseEvent & mouEvent)
	{
	};

	virtual void keyboardPress(const KeyboardEvent & keyEvent)
	{
	};

	virtual void keyboardRelease(const KeyboardEvent & keyEvent)
	{
	};

protected:

	virtual void bindTextures()
	{
		//////////////////////////////////////////////////////////////////////////		

		const Vector vecTexture(float((GetEnvironment().GetDataVolume().GetBlocksX()*BLOCKDIMENSION)),float((GetEnvironment().GetDataVolume().GetBlocksY()*BLOCKDIMENSION)),float((GetEnvironment().GetDataVolume().GetBlocksZ()*BLOCKDIMENSION)));
		const float vfPlaneX[] = {1.0f / vecTexture.GetX(), 0.0f, 0.0f, 0.5f*(1.0f / vecTexture.GetX())};
		const float vfPlaneY[] = {0.0f, 1.0f / vecTexture.GetY(), 0.0f, 0.5f*(1.0f / vecTexture.GetY())};
		const float vfPlaneZ[] = {0.0f, 0.0f, 1.0f / vecTexture.GetZ(), 0.5f*(1.0f / vecTexture.GetZ())};

		//////////////////////////////////////////////////////////////////////////

		glActiveTexture(GL_TEXTURE0);
		glBindTexture(GL_TEXTURE_2D,m_uVolumeColorTexture);

		glActiveTexture(GL_TEXTURE1);		
		glBindTexture(GL_TEXTURE_2D,m_uFrontColorTexture);

		glActiveTexture(GL_TEXTURE2);
		glBindTexture(GL_TEXTURE_2D,m_uBackColorTexture);

		glActiveTexture(GL_TEXTURE3);
		GetEnvironment().GetDataVolume().bind();
		glPushMatrix();
		glMultMatrixf(GetEnvironment().GetDataTransformation().Get());
		glTexGenfv(GL_S,GL_EYE_PLANE,vfPlaneX);
		glTexGenfv(GL_T,GL_EYE_PLANE,vfPlaneY);
		glTexGenfv(GL_R,GL_EYE_PLANE,vfPlaneZ);
		glPopMatrix();

		glActiveTexture(GL_TEXTURE4);
		GetEnvironment().GetSelectionVolume().bind();
		glPushMatrix();
		glMultMatrixf(GetEnvironment().GetDataTransformation().Get());
		glTexGenfv(GL_S,GL_EYE_PLANE,vfPlaneX);
		glTexGenfv(GL_T,GL_EYE_PLANE,vfPlaneY);
		glTexGenfv(GL_R,GL_EYE_PLANE,vfPlaneZ);
		glPopMatrix();

		glActiveTexture(GL_TEXTURE5);
		GetEnvironment().GetBackgroundSelectionColorTransferFunction().bind();

		glActiveTexture(GL_TEXTURE6);
		GetEnvironment().GetGhostSelectionColorTransferFunction().bind();

		glActiveTexture(GL_TEXTURE7);
		GetEnvironment().GetBackgroundLightingTransferFunction().bind();

		glActiveTexture(GL_TEXTURE8);
		GetEnvironment().GetGhostLightingTransferFunction().bind();

		glActiveTexture(GL_TEXTURE9);
		GetEnvironment().GetSelectionLightingTransferFunction().bind();

		//////////////////////////////////////////////////////////////////////////
	};

	virtual void releaseTextures()
	{
		glActiveTexture(GL_TEXTURE9);
		GetEnvironment().GetSelectionLightingTransferFunction().release();

		glActiveTexture(GL_TEXTURE8);
		GetEnvironment().GetGhostLightingTransferFunction().release();

		glActiveTexture(GL_TEXTURE7);
		GetEnvironment().GetBackgroundLightingTransferFunction().release();

		glActiveTexture(GL_TEXTURE6);
		GetEnvironment().GetGhostSelectionColorTransferFunction().release();

		glActiveTexture(GL_TEXTURE5);
		GetEnvironment().GetBackgroundSelectionColorTransferFunction().release();

		glActiveTexture(GL_TEXTURE4);
		GetEnvironment().GetSelectionVolume().release();

		glActiveTexture(GL_TEXTURE3);
		GetEnvironment().GetDataVolume().release();

		glActiveTexture(GL_TEXTURE2);
		glBindTexture(GL_TEXTURE_2D,0);

		glActiveTexture(GL_TEXTURE1);		
		glBindTexture(GL_TEXTURE_2D,0);

		glActiveTexture(GL_TEXTURE0);
		glBindTexture(GL_TEXTURE_2D,0);
	};

	virtual void renderBackgroundVolume()
	{
		const Vector vecTexture(float((GetEnvironment().GetDataVolume().GetBlocksX()*BLOCKDIMENSION)),float((GetEnvironment().GetDataVolume().GetBlocksY()*BLOCKDIMENSION)),float((GetEnvironment().GetDataVolume().GetBlocksZ()*BLOCKDIMENSION)));
		const float vfPlaneX[] = {1.0f / vecTexture.GetX(), 0.0f, 0.0f, 0.5f*(1.0f / vecTexture.GetX())};
		const float vfPlaneY[] = {0.0f, 1.0f / vecTexture.GetY(), 0.0f, 0.5f*(1.0f / vecTexture.GetY())};
		const float vfPlaneZ[] = {0.0f, 0.0f, 1.0f / vecTexture.GetZ(), 0.5f*(1.0f / vecTexture.GetZ())};

		glPushAttrib(GL_ALL_ATTRIB_BITS);

		glMatrixMode(GL_PROJECTION);
		glPushMatrix();
		glLoadIdentity();
		glMultMatrixf(GetEnvironment().GetProjectionTransformation().Get());

		glMatrixMode(GL_MODELVIEW);
		glPushMatrix();
		glLoadIdentity();
		glMultMatrixf(GetEnvironment().GetViewingTransformation().Get());
		glMultMatrixf(GetEnvironment().GetDataTransformation().Get());

		glActiveTexture(GL_TEXTURE0);
		glTexGenfv(GL_S,GL_EYE_PLANE,vfPlaneX);
		glTexGenfv(GL_T,GL_EYE_PLANE,vfPlaneY);
		glTexGenfv(GL_R,GL_EYE_PLANE,vfPlaneZ);

		m_proEntry.bind();

		//////////////////////////////////////////////////////////////////////////
		// Render back faces
		//////////////////////////////////////////////////////////////////////////
		glDrawBuffer(GL_COLOR_ATTACHMENT2_EXT);
		glColorMask(GL_TRUE,GL_TRUE,GL_TRUE,GL_TRUE);
		glClearColor(0.0f,0.0f,0.0f,0.0f);
		glClearDepth(0.0f);
		glDepthMask(GL_TRUE);
		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
		glEnable(GL_DEPTH_TEST);
		glEnable(GL_CULL_FACE);

		glColorMask(GL_TRUE,GL_TRUE,GL_TRUE,GL_TRUE);
		glDepthMask(GL_TRUE);
		glCullFace(GL_FRONT);
		glDepthFunc(GL_GREATER);
		renderBounds();

		//////////////////////////////////////////////////////////////////////////

		//////////////////////////////////////////////////////////////////////////
		// Render front faces
		//////////////////////////////////////////////////////////////////////////
		glDrawBuffer(GL_COLOR_ATTACHMENT1_EXT);
		glColorMask(GL_TRUE,GL_TRUE,GL_TRUE,GL_TRUE);
		glDepthMask(GL_TRUE);
		glClearColor(0.0f,0.0f,0.0f,0.0f);
		glClearDepth(1.0f);
		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
		glEnable(GL_DEPTH_TEST);
		glEnable(GL_CULL_FACE);

		glColorMask(GL_TRUE,GL_TRUE,GL_TRUE,GL_TRUE);
		glDepthMask(GL_TRUE);
		glCullFace(GL_BACK);
		glDepthFunc(GL_LESS);
		renderBounds();

		//////////////////////////////////////////////////////////////////////////

		m_proEntry.release();

		glDrawBuffer(GL_COLOR_ATTACHMENT0_EXT);

		glColorMask(GL_TRUE,GL_TRUE,GL_TRUE,GL_TRUE);
		glDepthMask(GL_FALSE);
		glClearColor(0.0f,0.0f,0.0f,0.0f);
		glClearDepth(0.0f);

		glEnable(GL_DEPTH_TEST);
		glDepthFunc(GL_LESS);		
		glDisable(GL_CULL_FACE);
		glCullFace(GL_FRONT);

		bindTextures();

		glMatrixMode(GL_PROJECTION);
		glLoadIdentity();

		glMatrixMode(GL_MODELVIEW);
		glLoadIdentity();

		const Vector vecView = (GetEnvironment().GetViewingTransformation().GetInverse() * Vector(0.0,0.0,1.0f)).GetNormalized();
		const Vector vecLight = (GetEnvironment().GetViewingTransformation().GetInverse() * Vector(-0.25,0.25,0.75f)).GetNormalized();
		const Vector vecHalf = (vecLight + vecView).GetNormalized();

		Matrix matInverseModelViewProjection = Matrix();
		Matrix matProjection = Matrix(GetEnvironment().GetProjectionTransformation().Get());
		Matrix matViewing = Matrix(GetEnvironment().GetViewingTransformation().Get());
		Matrix matModel = Matrix(GetEnvironment().GetDataTransformation().Get());	
		matInverseModelViewProjection *=  matProjection * matViewing * matModel;	
		//matInverseModelViewProjection.invert();

		m_proVolume.bind();

		glUniform1iARB(m_proVolume.GetUniformLocation("samImage"), 0);
		glUniform1iARB(m_proVolume.GetUniformLocation("samFront"), 1);
		glUniform1iARB(m_proVolume.GetUniformLocation("samBack"), 2);
		glUniform1iARB(m_proVolume.GetUniformLocation("samDataVolume"), 3);
		glUniform1iARB(m_proVolume.GetUniformLocation("samBackgroundSelectionColorTransferFunction"), 5);
		glUniform1iARB(m_proVolume.GetUniformLocation("samBackgroundLightingTransferFunction"), 7);
		glUniform3fvARB(m_proVolume.GetUniformLocation("vecLight"), 1, vecLight.Get());
		glUniform3fvARB(m_proVolume.GetUniformLocation("vecHalf"), 1, vecHalf.Get());
		glUniformMatrix4fvARB(m_proVolume.GetUniformLocation("matInverseModelViewProjection"),1,0,matInverseModelViewProjection.Get());
		glUniform1f(m_proVolume.GetUniformLocation("fTranslucencyFactor"),GetEnvironment().GetGhostGhosting()*32.0f);
		glUniform1f(m_proVolume.GetUniformLocation("fSoftnessFactor"),GetEnvironment().GetSelectionGhosting());
		glBegin(GL_QUADS);

		glTexCoord2f(0.0f,0.0f);
		glVertex3f(-1.0f,-1.0f,0.0f);

		glTexCoord2f(1.0f,0.0f);
		glVertex3f(1.0f,-1.0f,0.0f);

		glTexCoord2f(1.0f,1.0f);
		glVertex3f(1.0f,1.0f,0.0f);

		glTexCoord2f(0.0f,1.0f);
		glVertex3f(-1.0f,1.0f,0.0f);

		glEnd();

		m_proVolume.release();

		glMatrixMode(GL_PROJECTION);
		glPopMatrix();

		glMatrixMode(GL_MODELVIEW);
		glPopMatrix();

		releaseTextures();

		glPopAttrib();
	};


	void renderBounds()
	{
		Volume<DataVoxel>::Octree::Iterator iter(GetEnvironment().GetDataVolume().GetOctree());
		renderBounds(iter);
	};

	void renderBounds(const Volume<DataVoxel>::Octree::Iterator & iter)
	{
		if (GetEnvironment().GetBackgroundSelectionColorTransferFunction().IsVisble(0,(*iter).GetMinimum().GetValue(),(*iter).GetMaximum().GetValue()))
		{
			if (iter.IsLeaf())
			{
				const Vector & vecPosition((*iter).GetPosition());
				glTranslatef(vecPosition.GetX(),vecPosition.GetY(),vecPosition.GetZ());
				renderBlock();
				glTranslatef(-vecPosition.GetX(),-vecPosition.GetY(),-vecPosition.GetZ());
			}
			else
			{
				for (unsigned int i=0;i<8;i++)
					renderBounds(iter[i]);
			}
		}
	};

	virtual void renderBlock() const
	{
		const float fSizeX = float(BLOCKDIMENSION);
		const float fSizeY = float(BLOCKDIMENSION);
		const float fSizeZ = float(BLOCKDIMENSION);

		const float fStartX = -fSizeX*0.5f;
		const float fStartY = -fSizeY*0.5f;
		const float fStartZ = -fSizeZ*0.5f;

		const float fEndX = fStartX + fSizeX;
		const float fEndY = fStartY + fSizeY;
		const float fEndZ = fStartZ + fSizeZ;

		const float vfCube[2][4][3] =
		{
			// Left
			{
				{ fStartX,fEndY,fEndZ }, 
				{ fStartX,fEndY,fStartZ }, 
				{ fStartX,fStartY,fStartZ }, 
				{ fStartX,fStartY,fEndZ }, 
			},

			// Right
			{
				{ fEndX,fEndY,fStartZ }, 
				{ fEndX,fEndY,fEndZ }, 
				{ fEndX,fStartY,fEndZ }, 
				{ fEndX,fStartY,fStartZ }, 
			},
		};

		const float vfCubeStrips[10][3] =
		{
			{ fStartX,fEndY,fEndZ }, 
			{ fEndX,fEndY,fEndZ }, 

			{ fStartX,fEndY,fStartZ }, 
			{ fEndX,fEndY,fStartZ }, 

			{ fStartX,fStartY,fStartZ }, 
			{ fEndX,fStartY,fStartZ }, 

			{ fStartX,fStartY,fEndZ}, 
			{ fEndX,fStartY,fEndZ}, 

			{ fStartX,fEndY,fEndZ }, 
			{ fEndX,fEndY,fEndZ }, 
		};

		glBegin(GL_QUADS);

		glVertex3fv(vfCube[0][0]);
		glVertex3fv(vfCube[0][1]);
		glVertex3fv(vfCube[0][2]);
		glVertex3fv(vfCube[0][3]);

		glVertex3fv(vfCube[1][0]);
		glVertex3fv(vfCube[1][1]);
		glVertex3fv(vfCube[1][2]);
		glVertex3fv(vfCube[1][3]);

		glEnd();

		glBegin(GL_QUAD_STRIP);

		glVertex3fv(vfCubeStrips[0]);
		glVertex3fv(vfCubeStrips[1]);
		glVertex3fv(vfCubeStrips[2]);
		glVertex3fv(vfCubeStrips[3]);
		glVertex3fv(vfCubeStrips[4]);
		glVertex3fv(vfCubeStrips[5]);
		glVertex3fv(vfCubeStrips[6]);
		glVertex3fv(vfCubeStrips[7]);
		glVertex3fv(vfCubeStrips[8]);
		glVertex3fv(vfCubeStrips[9]);

		glEnd();
	};

	virtual void renderBox(const float fStartX, const float fStartY, const float fStartZ, const float fEndX, const float fEndY, const float fEndZ) const
	{
		const float vfCube[2][4][3] =
		{
			// Left
			{
				{ fStartX,fEndY,fEndZ }, 
				{ fStartX,fEndY,fStartZ }, 
				{ fStartX,fStartY,fStartZ }, 
				{ fStartX,fStartY,fEndZ }, 
			},

			// Right
			{
				{ fEndX,fEndY,fStartZ }, 
				{ fEndX,fEndY,fEndZ }, 
				{ fEndX,fStartY,fEndZ }, 
				{ fEndX,fStartY,fStartZ }, 
			},
		};

		const float vfCubeStrips[10][3] =
		{
			{ fStartX,fEndY,fEndZ }, 
			{ fEndX,fEndY,fEndZ }, 

			{ fStartX,fEndY,fStartZ }, 
			{ fEndX,fEndY,fStartZ }, 

			{ fStartX,fStartY,fStartZ }, 
			{ fEndX,fStartY,fStartZ }, 

			{ fStartX,fStartY,fEndZ}, 
			{ fEndX,fStartY,fEndZ}, 

			{ fStartX,fEndY,fEndZ }, 
			{ fEndX,fEndY,fEndZ }, 
		};

		glBegin(GL_QUADS);

		glVertex3fv(vfCube[0][0]);
		glVertex3fv(vfCube[0][1]);
		glVertex3fv(vfCube[0][2]);
		glVertex3fv(vfCube[0][3]);

		glVertex3fv(vfCube[1][0]);
		glVertex3fv(vfCube[1][1]);
		glVertex3fv(vfCube[1][2]);
		glVertex3fv(vfCube[1][3]);

		glEnd();

		glBegin(GL_QUAD_STRIP);

		glVertex3fv(vfCubeStrips[0]);
		glVertex3fv(vfCubeStrips[1]);
		glVertex3fv(vfCubeStrips[2]);
		glVertex3fv(vfCubeStrips[3]);
		glVertex3fv(vfCubeStrips[4]);
		glVertex3fv(vfCubeStrips[5]);
		glVertex3fv(vfCubeStrips[6]);
		glVertex3fv(vfCubeStrips[7]);
		glVertex3fv(vfCubeStrips[8]);
		glVertex3fv(vfCubeStrips[9]);

		glEnd();
	};


private:
	
	Environment & m_envEnvironment;
	unsigned int m_uWidth;
	unsigned int m_uHeight;

	unsigned int m_uVolumeFrameBuffer;

	unsigned int m_uFrontColorTexture;
	unsigned int m_uBackColorTexture;

	unsigned int m_uGeometryColorTexture;
	unsigned int m_uGeometryDepthTexture;

	unsigned int m_uVolumeColorTexture;
	unsigned int m_uVolumeDepthRenderBuffer;

	Program m_proVolume;
	Program m_proEntry;
};
#endif

