uniform sampler2D samImage;
uniform sampler2D samFront;
uniform sampler2D samBack;
uniform sampler3D samDataVolume;
uniform sampler3D samSelectionVolume;
uniform sampler2D samBackgroundSelectionColorTransferFunction;
uniform sampler2D samBackgroundLightingTransferFunction;
uniform vec3 vecLight;
uniform vec3 vecHalf;
uniform float fTranslucencyFactor;
uniform float fSoftnessFactor;

vec4 GetDataVolumeSample(const sampler3D samDataVolume, vec3 fPosition)
{
	/*
	vec4 vecSample;
	
	vecSample.a = texture3D(samDataVolume,fPosition).a;
	vecSample.x = texture3D(samDataVolume,fPosition-vec3(1.0/256.0,0.0,0.0)).a - vecSample.a;
	vecSample.y = texture3D(samDataVolume,fPosition-vec3(0.0,1.0/256.0,0.0)).a - vecSample.a;
	vecSample.z = texture3D(samDataVolume,fPosition-vec3(0.0,0.0,1.0/256.0)).a - vecSample.a;
	*/
	
	/*
	vecSample.a = texture3D(samDataVolume,fPosition).a;
	vecSample.x = texture3D(samDataVolume,fPosition-vec3(1.0/256.0,0.0,0.0)).a - texture3D(samDataVolume,fPosition+vec3(1.0/256.0,0.0,0.0)).a;
	vecSample.y = texture3D(samDataVolume,fPosition-vec3(0.0,1.0/256.0,0.0)).a - texture3D(samDataVolume,fPosition+vec3(0.0,1.0/256.0,0.0)).a;
	vecSample.z = texture3D(samDataVolume,fPosition-vec3(0.0,0.0,1.0/256.0)).a - texture3D(samDataVolume,fPosition+vec3(0.0,0.0,1.0/256.0)).a;

	return vecSample;
	*/
	
	return texture3D(samDataVolume,fPosition);
}

float GetSelectionVolumeSample(const sampler3D samSelectionVolume,  vec3 vecPosition)
{
	return texture3D(samSelectionVolume,vecPosition).a;
}

vec4 GetColor(const sampler2D samColorTransferFunction, vec2 vecSample)
{
	return texture2D(samColorTransferFunction,vecSample);
}

vec3 GetProducts(vec3 vecLight, vec3 vecHalf, vec3 vecGradient)
{
	vec3 vecProducts = vec3(0.0,0.0,0.0);
	float fLength = length(vecGradient);

	if (fLength > 0.0)
	{
		vec3 fNormal = normalize(vecGradient);
		vecProducts.x = dot(fNormal,/*normalize*/(vecLight));
		vecProducts.y = dot(fNormal,/*normalize*/(vecHalf));
		vecProducts.z = fLength;
	}

	return vecProducts;
}

vec4 GetLighting(const sampler2D samLightingTransferFunction, vec4 vecColor, vec3 vecProducts, float fDepth)
{
	vec4 vecLighting = vec4(0.0,0.0,0.0,1.0);
	vecLighting = texture2D(samLightingTransferFunction,(vecProducts.xy + 1.0)*0.5);
	
	float fShade = vecLighting.x+vecLighting.y+vecLighting.z;
	float fFactor = fTranslucencyFactor * fShade * (1.0 - fDepth) * (1.0 - vecColor.a);
	float fModulation = pow(vecProducts.z,pow(fFactor,fSoftnessFactor));
	vecLighting.a = fModulation;
	
	return vecLighting;
}

vec4 GetShaded(vec4 vecColor, vec4 vecLighting)
{
	vec4 vecShaded = vecColor;
	vecShaded.rgb = vecLighting.r + vecColor.rgb * vecLighting.g + vecLighting.b;
	vecShaded.a *= vecLighting.a;
	return vecShaded;
}

void main()
{
	//vec4 vecColor = texture2D(samImage,gl_TexCoord[0].xy);
	//discard(vecColor.a >= 1.0);

	vec4 vecFront = texture2D(samFront,gl_TexCoord[0].xy);
	vec4 vecBack = texture2D(samBack,gl_TexCoord[0].xy);
	
	vec3 vecDirection = vecBack.xyz - vecFront.xyz;
	float fDistance = length(vecDirection);
	
	if (fDistance == 0.0) discard;
	
	vecDirection = normalize(vecDirection);	
	vec4 vecRayColor = vec4(0.0,0.0,0.0,0.0);//vecColor;	
	
	vec4 vecRayIncrement;
	vecRayIncrement.xyz = vecDirection;
	vecRayIncrement.w = 1.0;
	vecRayIncrement /= 256.0;
	
	vec4 vecRayPosition;	
	vecRayPosition.xyz = vecFront.xyz;
	vecRayPosition.w = 0.0;
	
	do
	{
		vec4 vecSampleDataVolume = GetDataVolumeSample(samDataVolume,vecRayPosition.xyz );
		vec4 vecColorBackground = GetColor(samBackgroundSelectionColorTransferFunction,vec2(vecSampleDataVolume.a,0.0));
		
		if (vecColorBackground.a > 0.0)
		{			
			vec3 vecProducts = GetProducts(vecLight,vecHalf,vecSampleDataVolume.xyz);
			vec4 vecLighting = GetLighting(samBackgroundLightingTransferFunction,vecRayColor,vecProducts,vecRayPosition.w/fDistance);
			vec4 vecShaded = GetShaded(vecColorBackground,vecLighting);	
			
			vecShaded.a *= (1.0 - vecRayColor.a);
			vecShaded.rgb *= vecShaded.a;
			vecRayColor += vecShaded;
		}
		
		vecRayPosition += vecRayIncrement;				
	}			
	while (vecRayPosition.w < fDistance && vecRayColor.a < 1.0);
	
	gl_FragColor = vecRayColor;
}