#include "Explorer.h"

Explorer::Explorer(Environment & envEnvironment) : m_envEnvironment(envEnvironment), m_uWidth(32), m_uHeight(32),
	m_uVolumeFrameBuffer(0), m_uVolumeColorTexture(0), m_uVolumeDepthRenderBuffer(0), m_uFrontColorTexture(0), m_uBackColorTexture(0),
	m_proBackgroundDisplay(VertexShader("RendererVolumeCasterBackgroundDisplayVertex.glsl"),FragmentShader("RendererVolumeCasterBackgroundDisplayFragment.glsl")), m_proEntry(VertexShader("RendererVolumeCasterEntryVertex.glsl"),FragmentShader("RendererVolumeCasterEntryFragment.glsl")), disk(envEnvironment)
{
	
	m_vPlaneCoord = Vector(0.5,0.0,0.0);
	m_fPlaneScale = 0.5;
	
	m_vVolumeCoord = Vector(-0.5,0.0,0.0);
	m_fVolumeScale = 0.5;


	if (!GLEW_ARB_fragment_shader)
	{
		std::cerr << "GL_ARB_fragment_shader is not supported by your hardware or driver." << std::endl;
		throw Exception("GL_ARB_fragment_shader is not supported by your hardware or driver.");
	}

	if (!GLEW_ARB_vertex_shader)
	{
		std::cerr << "GL_ARB_fragment_shader is not supported by your hardware or driver." << std::endl;
		throw Exception("GL_ARB_fragment_shader is not supported by your hardware or driver.");
	}

	if (!GLEW_ARB_texture_non_power_of_two)
	{
		std::cerr << "GL_ARB_texture_non_power_of_two is not supported by your hardware or driver." << std::endl;
		throw Exception("GL_ARB_texture_non_power_of_two is not supported by your hardware or driver.");
	}

	if (!GLEW_ARB_texture_rectangle)
	{
		std::cerr << "GL_ARB_texture_rectangle is not supported by your hardware or driver." << std::endl;
		throw Exception("GL_ARB_texture_rectangle is not supported by your hardware or driver.");
	}		

	if (!GLEW_EXT_framebuffer_object)
	{
		std::cerr << "GL_EXT_framebuffer_object is not supported by your hardware or driver." << std::endl;
		throw Exception("GL_EXT_framebuffer_object is not supported by your hardware or driver.");
	}


	glGenFramebuffersEXT(1, &m_uVolumeFrameBuffer);

	glGenTextures(1,&m_uVolumeColorTexture);
	glGenTextures(1,&m_uFrontColorTexture);
	glGenTextures(1,&m_uBackColorTexture);

	glGenRenderbuffersEXT(1, &m_uVolumeDepthRenderBuffer);  
};

Explorer::~Explorer()
{
	if (m_uVolumeFrameBuffer)
		glDeleteFramebuffersEXT(1, &m_uVolumeFrameBuffer);
};

Environment & Explorer::GetEnvironment()
{
	return m_envEnvironment;
};

void Explorer::idle()
{
};

void Explorer::underlay()
{
};

void Explorer::display()
{	
	render(m_fVolumeScale,m_vVolumeCoord.GetX());

	glPushMatrix();
	glTranslatef(m_vPlaneCoord.GetX(),0.0,0.0);
	glScalef(m_fPlaneScale,m_fPlaneScale,1.0);
	glColor3f(0.0,0.0,0.0);
	glBegin(GL_LINES);
	glVertex3f(-1.0,-1.0,0.0);
	glVertex3f(1.0,-1.0,0.0);
	glVertex3f(1.0,-1.0,0.0);
	glVertex3f(1.0,1.0,0.0);
	glVertex3f(1.0,1.0,0.0);
	glVertex3f(-1.0,1.0,0.0);
	glVertex3f(-1.0,1.0,0.0);
	glVertex3f(-1.0,-1.0,0.0);
	glEnd();
	disk.setFocus(0,true);
	disk.display();
	glPopMatrix();

	
};

void Explorer::overlay()
{
};

void Explorer::reshape(const unsigned int uWidth, const unsigned int uHeight)
{
	GLenum glStatus;

	m_uWidth = uWidth;
	m_uHeight = uHeight;

	//////////////////////////////////////////////////////////////////////////
	glBindTexture(GL_TEXTURE_2D,m_uVolumeColorTexture);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_CLAMP);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_CLAMP);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
	glTexImage2D(GL_TEXTURE_2D,0,GL_RGBA16F_ARB,uWidth,uHeight,0,GL_RGBA,GL_UNSIGNED_BYTE,NULL);
	glBindTexture(GL_TEXTURE_2D,0);
	//////////////////////////////////////////////////////////////////////////
	glBindTexture(GL_TEXTURE_2D,m_uFrontColorTexture);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_CLAMP);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_CLAMP);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
	glTexImage2D(GL_TEXTURE_2D,0,GL_RGBA16F_ARB,uWidth,uHeight,0,GL_RGBA,GL_UNSIGNED_BYTE,NULL);
	glBindTexture(GL_TEXTURE_2D,0);
	//////////////////////////////////////////////////////////////////////////
	glBindTexture(GL_TEXTURE_2D,m_uBackColorTexture);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_CLAMP);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_CLAMP);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
	glTexImage2D(GL_TEXTURE_2D,0,GL_RGBA16F_ARB,uWidth,uHeight,0,GL_RGBA,GL_UNSIGNED_BYTE,NULL);
	glBindTexture(GL_TEXTURE_2D,0);
	//////////////////////////////////////////////////////////////////////////
	glBindRenderbufferEXT(GL_RENDERBUFFER_EXT, m_uVolumeDepthRenderBuffer);
	glRenderbufferStorageEXT(GL_RENDERBUFFER_EXT,GL_DEPTH_COMPONENT, uWidth, uHeight);  
	glBindRenderbufferEXT(GL_RENDERBUFFER_EXT, 0);
	//////////////////////////////////////////////////////////////////////////
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, m_uVolumeFrameBuffer);
	glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT,GL_COLOR_ATTACHMENT0_EXT,GL_TEXTURE_2D, m_uVolumeColorTexture, 0);
	glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT,GL_COLOR_ATTACHMENT1_EXT,GL_TEXTURE_2D, m_uFrontColorTexture, 0);
	glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT,GL_COLOR_ATTACHMENT2_EXT,GL_TEXTURE_2D, m_uBackColorTexture, 0);
	glFramebufferRenderbufferEXT(GL_FRAMEBUFFER_EXT,GL_DEPTH_ATTACHMENT_EXT,GL_RENDERBUFFER_EXT, m_uVolumeDepthRenderBuffer);  

	glStatus = glCheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT);
	if (glStatus  != GL_FRAMEBUFFER_COMPLETE_EXT)
	{
		std::cerr << "Error initializing volume framebuffer." << std::endl << std::endl;
		throw Exception("Error initializing volume framebuffer.");
	}

	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);
	//////////////////////////////////////////////////////////////////////////
};


void Explorer::mousePress(const MouseEvent & mouEvent)
{
	Vector v = mouEvent.GetPosition();
	if (v.GetX()>0 && v.GetY()<0.5 && v.GetY()>-0.5) {
		disk.mousePress(mouEvent);
		Vector * texCoords = new Vector[4];
		texCoords = disk.getTextureCoords(0);

		int sizex = m_envEnvironment.GetDataVolume().GetSizeX();
		int sizey = m_envEnvironment.GetDataVolume().GetSizeY();
		int sizez = m_envEnvironment.GetDataVolume().GetSizeZ();

		float x = v.GetX(); 
		float y = v.GetY() + 0.5;
		float z = texCoords[0].GetY();

		//x = 1 - x;	
		//y = 1 - y;


		m_vCoord.SetX(x*sizex);
		m_vCoord.SetY(y*sizey);
		m_vCoord.SetZ(z*sizez);
		float threshold =  m_envEnvironment.GetDataVolume().Get((unsigned int)m_vCoord.GetX(),(unsigned int)m_vCoord.GetY(),(unsigned int)m_vCoord.GetZ()).GetValue();
		std::cout << m_vCoord << std::endl;
		std::cout << threshold << std::endl;
		m_envEnvironment.SetCutAwayThreshold(threshold);


	}

	
};

void Explorer::mouseRelease(const MouseEvent & mouEvent)
{	
	disk.mouseRelease(mouEvent);
};

void Explorer::mouseMove(const MouseEvent & mouEvent)
{
	disk.mouseMove(mouEvent);
};

void Explorer::keyboardPress(const KeyboardEvent & keyEvent)
{
};

void Explorer::keyboardRelease(const KeyboardEvent & keyEvent)
{
};


void Explorer::bindTextures()
{
	//////////////////////////////////////////////////////////////////////////		

	const Vector vecTexture(float((GetEnvironment().GetDataVolume().GetBlocksX()*BLOCKDIMENSION)),float((GetEnvironment().GetDataVolume().GetBlocksY()*BLOCKDIMENSION)),float((GetEnvironment().GetDataVolume().GetBlocksZ()*BLOCKDIMENSION)));
	const float vfPlaneX[] = {1.0f / vecTexture.GetX(), 0.0f, 0.0f, 0.5f*(1.0f / vecTexture.GetX())};
	const float vfPlaneY[] = {0.0f, 1.0f / vecTexture.GetY(), 0.0f, 0.5f*(1.0f / vecTexture.GetY())};
	const float vfPlaneZ[] = {0.0f, 0.0f, 1.0f / vecTexture.GetZ(), 0.5f*(1.0f / vecTexture.GetZ())};

	//////////////////////////////////////////////////////////////////////////

	// Set up geometry depth texture
	glActiveTexture(GL_TEXTURE0);		
	glBindTexture(GL_TEXTURE_2D,m_uFrontColorTexture);

	// Set up geometry depth texture
	glActiveTexture(GL_TEXTURE1);
	glBindTexture(GL_TEXTURE_2D,m_uBackColorTexture);

	// Set up volume texture
	glActiveTexture(GL_TEXTURE2);
	GetEnvironment().GetDataVolume().bind();
	glPushMatrix();
	glMultMatrixf(GetEnvironment().GetDataTransformation().Get());
	glTexGenfv(GL_S,GL_EYE_PLANE,vfPlaneX);
	glTexGenfv(GL_T,GL_EYE_PLANE,vfPlaneY);
	glTexGenfv(GL_R,GL_EYE_PLANE,vfPlaneZ);
	glPopMatrix();

	// Set up background/selection scalar transfer function texture
	glActiveTexture(GL_TEXTURE3);
	GetEnvironment().GetBackgroundSelectionColorTransferFunction().bind();

	// Set up background lighting transfer function
	glActiveTexture(GL_TEXTURE4);
	GetEnvironment().GetBackgroundLightingTransferFunction().bind();

	//////////////////////////////////////////////////////////////////////////
};

void Explorer::releaseTextures()
{
	// Set up geometry depth texture
	glActiveTexture(GL_TEXTURE1);
	glBindTexture(GL_TEXTURE_2D,0);

	// Set up geometry depth texture
	glActiveTexture(GL_TEXTURE0);		
	glBindTexture(GL_TEXTURE_2D,0);

};
void Explorer::render(float scale, float pos) {
	glPushAttrib(GL_ALL_ATTRIB_BITS);
	
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, m_uVolumeFrameBuffer);
	renderVolume();

	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);

	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();

	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();

	glActiveTexture(GL_TEXTURE0);

	glEnable(GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D,m_uVolumeColorTexture);
	
	glTranslatef(pos,0.0,0.0);
	glScalef(scale,scale,1.0);
	glBegin(GL_QUADS);

	glTexCoord2f(0.0f,0.0f);
	glVertex2f(-1.0f,-1.0f);

	glTexCoord2f(1.0f,0.0f);
	glVertex2f(1.0f,-1.0f);

	glTexCoord2f(1.0f,1.0f);
	glVertex2f(1.0f,1.0f);

	glTexCoord2f(0.0f,1.0f);
	glVertex2f(-1.0f,1.0f);

	glEnd();

	glBindTexture(GL_TEXTURE_2D,0);

	glMatrixMode(GL_PROJECTION);
	glPopMatrix();

	glMatrixMode(GL_MODELVIEW);
	glPopMatrix();
	
	glPopAttrib();

};


void Explorer::renderVolume()
{
	glPushAttrib(GL_ALL_ATTRIB_BITS);

	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	glMultMatrixf(GetEnvironment().GetProjectionTransformation().Get());

	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();
	glMultMatrixf(GetEnvironment().GetViewingTransformation().Get());
	glMultMatrixf(GetEnvironment().GetDataTransformation().Get());
	Volume<DataVoxel>::Octree::Iterator iter(GetEnvironment().GetDataVolume().GetOctree());

	const Vector vecTexture(float((GetEnvironment().GetDataVolume().GetBlocksX()*BLOCKDIMENSION)),float((GetEnvironment().GetDataVolume().GetBlocksY()*BLOCKDIMENSION)),float((GetEnvironment().GetDataVolume().GetBlocksZ()*BLOCKDIMENSION)));
	const float vfPlaneX[] = {1.0f / vecTexture.GetX(), 0.0f, 0.0f, 0.5f*(1.0f / vecTexture.GetX())};
	const float vfPlaneY[] = {0.0f, 1.0f / vecTexture.GetY(), 0.0f, 0.5f*(1.0f / vecTexture.GetY())};
	const float vfPlaneZ[] = {0.0f, 0.0f, 1.0f / vecTexture.GetZ(), 0.5f*(1.0f / vecTexture.GetZ())};

	glActiveTexture(GL_TEXTURE0);
	glTexGenfv(GL_S,GL_EYE_PLANE,vfPlaneX);
	glTexGenfv(GL_T,GL_EYE_PLANE,vfPlaneY);
	glTexGenfv(GL_R,GL_EYE_PLANE,vfPlaneZ);

	m_proEntry.bind();

	//////////////////////////////////////////////////////////////////////////
	// Render back faces
	//////////////////////////////////////////////////////////////////////////
	glDrawBuffer(GL_COLOR_ATTACHMENT2_EXT);

	glColorMask(GL_TRUE,GL_TRUE,GL_TRUE,GL_TRUE);
	glDepthMask(GL_TRUE);
	glDepthFunc(GL_GREATER);
	glEnable(GL_DEPTH_TEST);
	glCullFace(GL_FRONT);
	glEnable(GL_CULL_FACE);
	glClearDepth(0.0f);		
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	glColor4f(1.0f,0.0f,0.0f,1.0f);
	renderBounds(iter);

	
	/*
	glCullFace(GL_FRONT);
	glEnable(GL_CULL_FACE);

	glDepthFunc(GL_LESS);
	glTranslatef(50,50,50);
	renderBox(100,100,30);
	glTranslatef(-50,-50,-50);
	*/



	//////////////////////////////////////////////////////////////////////////

	//////////////////////////////////////////////////////////////////////////
	// Render front faces
	//////////////////////////////////////////////////////////////////////////
	glDrawBuffer(GL_COLOR_ATTACHMENT1_EXT);

	glColorMask(GL_TRUE,GL_TRUE,GL_TRUE,GL_TRUE);
	glDepthMask(GL_TRUE);
	glDepthFunc(GL_LESS);
	glEnable(GL_DEPTH_TEST);
	glCullFace(GL_BACK);
	glEnable(GL_CULL_FACE);
	glClearDepth(1.0f);		
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	glColor4f(0.0f,1.0f,0.0f,1.0f);
	renderBounds(iter);


	//
	/*
	glCullFace(GL_BACK);
	glEnable(GL_CULL_FACE);

	glDepthFunc(GL_GREATER);
	glTranslatef(50,50,50);
	renderBox(100,100,30);
	glTranslatef(-50,-50,-50);
	*/
	//


	//////////////////////////////////////////////////////////////////////////

	m_proEntry.release();


	glDrawBuffer(GL_COLOR_ATTACHMENT0_EXT);		
	glColorMask(GL_TRUE,GL_TRUE,GL_TRUE,GL_TRUE);
	glDepthMask(GL_FALSE);
	glDepthFunc(GL_GREATER);
	glEnable(GL_DEPTH_TEST);	
	glCullFace(GL_FRONT);
	glDisable(GL_CULL_FACE);
	glClear(GL_COLOR_BUFFER_BIT);
	bindTextures();
	
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	const Vector vecView = (GetEnvironment().GetViewingTransformation().GetInverse() * Vector(0.0,0.0,1.0f)).GetNormalized();
	const Vector vecLight = (GetEnvironment().GetViewingTransformation().GetInverse() * Vector(-0.25,0.25,0.75f)).GetNormalized();
	const Vector vecHalf = (vecLight + vecView).GetNormalized();
	
	//Added by Asmund Birkeland		
	float threshold = 0.0;
	bool cutaway = GetEnvironment().GetCutAwayEnabled();
	
	if (cutaway) {
		threshold = GetEnvironment().GetCutAwayThreshold();
	}
	float * CA = new float[2];
	CA[0] = (float)cutaway;
	CA[1] = threshold;
	//end Birkeland

	m_proBackgroundDisplay.bind();

	glUniform1iARB(m_proBackgroundDisplay.GetUniformLocation("samFront"), 0);
	glUniform1iARB(m_proBackgroundDisplay.GetUniformLocation("samBack"), 1);
	glUniform1iARB(m_proBackgroundDisplay.GetUniformLocation("samDataVolume"), 2);
	glUniform1iARB(m_proBackgroundDisplay.GetUniformLocation("samBackgroundSelectionColorTransferFunction"), 3);
	glUniform1iARB(m_proBackgroundDisplay.GetUniformLocation("samBackgroundLightingTransferFunction"), 4);
	glUniform3fvARB(m_proBackgroundDisplay.GetUniformLocation("vecLight"), 1, vecLight.Get());
	glUniform3fvARB(m_proBackgroundDisplay.GetUniformLocation("vecHalf"), 1, vecHalf.Get());

	//Added by Asmund Birkeland
	glUniform2fvARB(m_proBackgroundDisplay.GetUniformLocation("cutaway"),1,CA);
	
	glBegin(GL_QUADS);
	
	glTexCoord2f(0.0f,0.0f);
	glVertex3f(-1.0f,-1.0f,1.0f);

	glTexCoord2f(1.0f,0.0f);
	glVertex3f(1.0f,-1.0f,1.0f);

	glTexCoord2f(1.0f,1.0f);
	glVertex3f(1.0f,1.0f,1.0f);

	glTexCoord2f(0.0f,1.0f);
	glVertex3f(-1.0f,1.0f,1.0f);

	glEnd();
	
	m_proBackgroundDisplay.release();
	
	glMatrixMode(GL_PROJECTION);
	glPopMatrix();

	glMatrixMode(GL_MODELVIEW);
	glPopMatrix();

	releaseTextures();
	
	glPopAttrib();
};

void Explorer::renderBounds(const Volume<DataVoxel>::Octree::Iterator & iter)
{
	if (GetEnvironment().GetBackgroundSelectionColorTransferFunction().IsVisble(0,(*iter).GetMinimum().GetValue(),(*iter).GetMaximum().GetValue()))
	{
		if (iter.IsLeaf())
		{
			const Vector & vecPosition((*iter).GetPosition());
			glTranslatef(vecPosition.GetX(),vecPosition.GetY(),vecPosition.GetZ());
			renderBlock();
			glTranslatef(-vecPosition.GetX(),-vecPosition.GetY(),-vecPosition.GetZ());
		}
		else
		{
			for (unsigned int i=0;i<8;i++)
				renderBounds(iter[i]);
		}
	}
};

void Explorer::renderBlock() const
{
	const float fSizeX = float(BLOCKDIMENSION);
	const float fSizeY = float(BLOCKDIMENSION);
	const float fSizeZ = float(BLOCKDIMENSION);

	const float fStartX = -fSizeX*0.5f;
	const float fStartY = -fSizeY*0.5f;
	const float fStartZ = -fSizeZ*0.5f;

	const float fEndX = fStartX + fSizeX;
	const float fEndY = fStartY + fSizeY;
	const float fEndZ = fStartZ + fSizeZ;

	const float vfCube[2][4][3] =
	{
		// Left
		{
			{ fStartX,fEndY,fEndZ }, 
			{ fStartX,fEndY,fStartZ }, 
			{ fStartX,fStartY,fStartZ }, 
			{ fStartX,fStartY,fEndZ }, 
		},

		// Right
		{
			{ fEndX,fEndY,fStartZ }, 
			{ fEndX,fEndY,fEndZ }, 
			{ fEndX,fStartY,fEndZ }, 
			{ fEndX,fStartY,fStartZ }, 
		},
	};

	const float vfCubeStrips[10][3] =
	{
		{ fStartX,fEndY,fEndZ }, 
		{ fEndX,fEndY,fEndZ }, 

		{ fStartX,fEndY,fStartZ }, 
		{ fEndX,fEndY,fStartZ }, 

		{ fStartX,fStartY,fStartZ }, 
		{ fEndX,fStartY,fStartZ }, 

		{ fStartX,fStartY,fEndZ}, 
		{ fEndX,fStartY,fEndZ}, 

		{ fStartX,fEndY,fEndZ }, 
		{ fEndX,fEndY,fEndZ }, 
	};

	glBegin(GL_QUADS);

	glVertex3fv(vfCube[0][0]);
	glVertex3fv(vfCube[0][1]);
	glVertex3fv(vfCube[0][2]);
	glVertex3fv(vfCube[0][3]);

	glVertex3fv(vfCube[1][0]);
	glVertex3fv(vfCube[1][1]);
	glVertex3fv(vfCube[1][2]);
	glVertex3fv(vfCube[1][3]);

	glEnd();

	glBegin(GL_QUAD_STRIP);

	glVertex3fv(vfCubeStrips[0]);
	glVertex3fv(vfCubeStrips[1]);
	glVertex3fv(vfCubeStrips[2]);
	glVertex3fv(vfCubeStrips[3]);
	glVertex3fv(vfCubeStrips[4]);
	glVertex3fv(vfCubeStrips[5]);
	glVertex3fv(vfCubeStrips[6]);
	glVertex3fv(vfCubeStrips[7]);
	glVertex3fv(vfCubeStrips[8]);
	glVertex3fv(vfCubeStrips[9]);

	glEnd();
};

void Explorer::renderBox(const float fSizeX, const float fSizeY, const float fSizeZ) const
{
	const float fStartX = -fSizeX*0.5f;
	const float fStartY = -fSizeY*0.5f;
	const float fStartZ = -fSizeZ*0.5f;

	const float fEndX = fStartX + fSizeX;
	const float fEndY = fStartY + fSizeY;
	const float fEndZ = fStartZ + fSizeZ;

	const float vfCube[2][4][3] =
	{
		// Left
		{
			{ fStartX,fEndY,fEndZ }, 
			{ fStartX,fEndY,fStartZ }, 
			{ fStartX,fStartY,fStartZ }, 
			{ fStartX,fStartY,fEndZ }, 
		},

		// Right
		{
			{ fEndX,fEndY,fStartZ }, 
			{ fEndX,fEndY,fEndZ }, 
			{ fEndX,fStartY,fEndZ }, 
			{ fEndX,fStartY,fStartZ }, 
		},
	};

	const float vfCubeStrips[10][3] =
	{
		{ fStartX,fEndY,fEndZ }, 
		{ fEndX,fEndY,fEndZ }, 

		{ fStartX,fEndY,fStartZ }, 
		{ fEndX,fEndY,fStartZ }, 

		{ fStartX,fStartY,fStartZ }, 
		{ fEndX,fStartY,fStartZ }, 

		{ fStartX,fStartY,fEndZ}, 
		{ fEndX,fStartY,fEndZ}, 

		{ fStartX,fEndY,fEndZ }, 
		{ fEndX,fEndY,fEndZ }, 
	};

	glBegin(GL_QUADS);

	glVertex3fv(vfCube[0][0]);
	glVertex3fv(vfCube[0][1]);
	glVertex3fv(vfCube[0][2]);
	glVertex3fv(vfCube[0][3]);

	glVertex3fv(vfCube[1][0]);
	glVertex3fv(vfCube[1][1]);
	glVertex3fv(vfCube[1][2]);
	glVertex3fv(vfCube[1][3]);

	glEnd();

	glBegin(GL_QUAD_STRIP);

	glVertex3fv(vfCubeStrips[0]);
	glVertex3fv(vfCubeStrips[1]);
	glVertex3fv(vfCubeStrips[2]);
	glVertex3fv(vfCubeStrips[3]);
	glVertex3fv(vfCubeStrips[4]);
	glVertex3fv(vfCubeStrips[5]);
	glVertex3fv(vfCubeStrips[6]);
	glVertex3fv(vfCubeStrips[7]);
	glVertex3fv(vfCubeStrips[8]);
	glVertex3fv(vfCubeStrips[9]);

	glEnd();
};


