#ifndef __PLUGIN_RENDERER_EXPLORER_EXPLOERER_H
#define __PLUGIN_RENDERER_EXPLORER_EXPLOERER_H

#include "volumeshop.h"

#include "Environment.h"
#include "Volume.h"
#include "Renderer.h"
#include "Vector.h"
#include "Quaternion.h"
#include "Matrix.h"
#include "Shader.h"
#include "Color.h"
#include "Timer.h"
#include "Label.h"
#include "ColorTransferFunction.h"
#include "LightingTransferFunction.h"
#include "Exception.h"


#include "../plugin_renderer_DiskRenderer/DiskRender.h"

class Explorer: public Renderer
{
public:
	Explorer(Environment & envEnvironment);
	
	~Explorer();

	Environment & GetEnvironment();

	void idle();

	void underlay();

	void display();

	void overlay();

	void reshape(const unsigned int uWidth, const unsigned int uHeight);

	void mousePress(const MouseEvent & mouEvent);

	void mouseRelease(const MouseEvent & mouEvent);

	void mouseMove(const MouseEvent & mouEvent);

	void keyboardPress(const KeyboardEvent & keyEvent);

	void keyboardRelease(const KeyboardEvent & keyEvent);
		

	void bindTextures();

	void releaseTextures();

	void renderVolume();

	void renderBounds(const Volume<DataVoxel>::Octree::Iterator & iter);

	void renderBlock() const;

	void renderBox(const float fSizeX, const float fSizeY, const float fSizeZ) const;

	void render(float scale, float pos);

private:
	DiskRender disk;
	Environment & m_envEnvironment;

	Vector m_vCoord;
	Vector m_vPlaneCoord;
	Vector m_vVolumeCoord;
	float m_fPlaneScale;
	float m_fVolumeScale;



	unsigned int m_uWidth;
	unsigned int m_uHeight;

	unsigned int m_uVolumeFrameBuffer;

	unsigned int m_uFrontColorTexture;
	unsigned int m_uBackColorTexture;

	unsigned int m_uGeometryColorTexture;
	unsigned int m_uGeometryDepthTexture;

	unsigned int m_uVolumeColorTexture;
	unsigned int m_uVolumeDepthRenderBuffer;

	Program m_proBackgroundDisplay;
	Program m_proEntry;

};
#endif
