#include "DiskRender.h"

DiskRender::DiskRender(Environment & envEnvironment): m_envEnvironment(envEnvironment), m_shaderProgram(FragmentShader("DiskRenderShader.glsl")){
		m_fXY[0] = m_fYZ[0] = m_fXZ[0] = 0.25;
		m_iQuadrant = m_iOldQuadrant = -1;
		m_bButton2 = m_bFocus = false;
		m_fZPlane = 0.0;
		
		//generate the different planes
		m_pPlane[0] = Plane(Vector(-0.5f,-0.5f,0.0f), 1.0);
		m_pPlane[1] = Plane(Vector(-0.5f,0.5f,0.0f), 1.0);
		m_pPlane[2] = Plane(Vector(0.5f,0.5f,0.0f), 1.0);
		m_pPlane[3] = Plane(Vector(0.5f,-0.5f,0.0f), 1.0);
		
};

DiskRender::~DiskRender(){};

void DiskRender::idle(){};
	
void DiskRender::underlay(){};
	
void DiskRender::display(){
	
	
	glEnable(GL_BLEND);

	glEnable(GL_DEPTH_TEST);
	glPolygonMode(GL_FRONT,GL_FILL);

	
	
	//Generate the different texture coords and putting them into each plane
	genTexXZ(m_pPlane[0].GetTextureCoords());
	genTexYZ(m_pPlane[1].GetTextureCoords());
	genTexXY(m_pPlane[2].GetTextureCoords());
	genTexARB(m_pPlane[3].GetTextureCoords());
	

	//draw a single large selected plane
	if (m_bFocus) {	
		glPushMatrix();
		glScalef(2,2,2);
		

		m_pPlane[m_iQuadrant].render(m_envEnvironment, &m_shaderProgram);
	

		glPopMatrix();
	}
	else {
		
		//draw all the planes cascaded
		for (int i = 0; i < 4; i++) {
			Vector p = m_pPlane[i].GetPosition();
			float s = m_pPlane[i].GetScale();
			glPushMatrix();
			glTranslatef(p.GetX(),p.GetY(),p.GetZ());
			glScalef(s,s,s);
			m_pPlane[i].render(m_envEnvironment, &m_shaderProgram);
			glPopMatrix();
		}
		
		glPushMatrix();
		renderBorder();
		glPopMatrix();
	}


	glEnable(GL_BLEND);
	glDisable(GL_DEPTH_TEST);


};

void DiskRender::overlay(){};

void DiskRender::reshape(const unsigned int uWidth, const unsigned int uHeight){};

//this is self-explanatory
void DiskRender::mousePress(const MouseEvent & mouEvent){
	
	m_vStart = mouEvent.GetPosition();
	
	if (!m_bFocus) {m_iQuadrant = GetPlane(m_vStart);}
	
	if (mouEvent.GetButton() == 0) {m_bButton0=true;}
	
	if (mouEvent.GetButton() == 2) {m_bButton2=true;}
	
	if (m_bButton0 && m_bButton2) {m_bFocus = !m_bFocus;}
	
	
};

//store all the values of all the planes and the rotation of the arbituary plane
void DiskRender::mouseRelease(const MouseEvent & mouEvent){
	
	m_vOld.SetX(m_fRotX);
	m_vOld.SetY(m_fRotY);
	m_vOld.SetZ(m_fZPlane);

	m_fXZ[1] = m_fXZ[0];
	m_fXY[1] = m_fXY[0];
	m_fYZ[1] = m_fYZ[0];

	if (mouEvent.GetButton() == 0) {m_bButton0=false;}
	
	if (mouEvent.GetButton() == 2) {m_bButton2=false;}
	
	
};

//update the different cutting depths and rotation
void DiskRender::mouseMove(const MouseEvent & mouEvent){
	
	float xPos = mouEvent.GetPosition().GetX();
	float yPos = mouEvent.GetPosition().GetY();
	
	if (m_bButton0) {
		float diff = yPos-m_vStart.GetY();
		switch (m_iQuadrant) {
			case 0:
				m_fXZ[0] = diff + m_fXZ[1];
				break;
			case 1:
				m_fYZ[0] = diff + m_fYZ[1];
				break;
			case 2:
				m_fXY[0] = diff + m_fXY[1];
				break;
			case 3:
				m_fZPlane = diff + m_vOld.GetZ();	
		}
	}
	if (m_bButton2) {
		float diffX = xPos-m_vStart.GetX();
		float diffY = yPos-m_vStart.GetY();
		m_fRotX = 100*diffX + m_vOld.GetX();
		m_fRotY = 100*diffY + m_vOld.GetY();
	}
};

void DiskRender::keyboardPress(const KeyboardEvent & keyEvent){};	
void DiskRender::keyboardRelease(const KeyboardEvent & keyEvent){};

//returns the index for the plane selected at mousepress
int DiskRender::GetPlane(Vector mousePos) {

	for (int i = 0; i < 4; i++) {
		if ((mousePos.GetX() < (m_pPlane[i].GetPosition().GetX() + m_pPlane[i].GetScale()/2)) && (mousePos.GetX() > (m_pPlane[i].GetPosition().GetX() - m_pPlane[i].GetScale()/2))
			&& (mousePos.GetY() < (m_pPlane[i].GetPosition().GetY() + m_pPlane[i].GetScale()/2)) && (mousePos.GetY() > (m_pPlane[i].GetPosition().GetY() - m_pPlane[i].GetScale()/2)))
		{
			return i;
		}
	}
	return -1;
}

//rotates a vector where degrees are the degrees of rotation, and direction an integer where 1 is x axis and 2 is y axis
void DiskRender::Rotate(double degrees, int direction, Vector * v) {
	
	degrees = degrees*PI/180;
	double matrixX[3][3] = 
	{
	{1,	0,				0},
	{0,	cos(degrees),	-sin(degrees)},
	{0,	sin(degrees),	cos(degrees)},
	};

	double matrixY[3][3] = 
	{
	{cos(degrees),	0,	-sin(degrees)},
	{0,				1,	0},
	{sin(degrees),	0,	cos(degrees)},
	};

	switch (direction) {
		case 1:
			matrixMul(matrixX,v);
			break;
		case 2:
			matrixMul(matrixY,v);
	}
};

//multplies a matrix with a vector
void DiskRender::matrixMul(double matrix[][3] , Vector * v) {
	
	double vec[3] =  {v->GetX(),v->GetY(),v->GetZ()};	
	double temp[3] = {0,0,0};

	for (int i = 0; i < 3 ; i++) {
		for (int j = 0; j < 3 ; j++) {
			temp[i] += matrix[i][j] * vec[j];
		}
	}
	v->SetX((float)temp[0]);
	v->SetY((float)temp[1]);
	v->SetZ((float)temp[2]);
};	

//generate texturecoords that cut through the XY plane
Vector * DiskRender::genTexXY(Vector * texCoord) {
	
	texCoord[0] = Vector(0.0,0.0,2*m_fXY[0]);
	texCoord[1] = Vector(1.0,0.0,2*m_fXY[0]);
	texCoord[2] = Vector(1.0,1.0,2*m_fXY[0]);
	texCoord[3] = Vector(0.0,1.0,2*m_fXY[0]);
	
	return texCoord;
};

//generate texturecoords that cut through the XZ plane
Vector * DiskRender::genTexXZ(Vector * texCoord) {
	
	texCoord[0] = Vector(0.0,2*m_fXZ[0],0.0);
	texCoord[1] = Vector(1.0,2*m_fXZ[0],0.0);
	texCoord[2] = Vector(1.0,2*m_fXZ[0],1.0);
	texCoord[3] = Vector(0.0,2*m_fXZ[0],1.0);
	
	return texCoord;
};

//generate texturecoords that cut through the YZ plane
Vector * DiskRender::genTexYZ(Vector * texCoord) {
	
	texCoord[0] = Vector(2*m_fYZ[0],0.0,0.0);
	texCoord[1] = Vector(2*m_fYZ[0],0.0,1.0);
	texCoord[2] = Vector(2*m_fYZ[0],1.0,1.0);
	texCoord[3] = Vector(2*m_fYZ[0],1.0,0.0);
	
	return texCoord;
};

//generate texturecoords that cut through the arbitrary plane
Vector * DiskRender::genTexARB(Vector * texCoord) {
	
	texCoord[0] = Vector(-0.5,-0.5,m_fZPlane);
	texCoord[1] = Vector(0.5,-0.5,m_fZPlane);
	texCoord[2] = Vector(0.5,0.5,m_fZPlane);
	texCoord[3] = Vector(-0.5,0.5,m_fZPlane);
	
	for (int i = 0; i<4; i++) {
		Rotate(m_fRotX,2,&texCoord[i]);
		Rotate(m_fRotY,1,&texCoord[i]);
		texCoord[i] += Vector(0.5,0.5,0.5);
	}
	return texCoord;
};

//renders the borders for the cascaded planes
void DiskRender::renderBorder(){

	glColor3f(0.5,0.5,0.5);
	glBegin(GL_QUADS);
	glVertex3f(-0.01f,-1.0f,-0.4f);
	glVertex3f(0.01f,-1.0f,-0.4f);
	glVertex3f(0.01f,1.0f,-0.4f);
	glVertex3f(-0.01f,1.0f,-0.4f);
	glEnd();
	
	glRotatef(90,0,0,1);
	glTranslatef(0.0,-0.5,0.0);
	glBegin(GL_QUADS);
	glVertex3f(-0.01f,-1.0f,-0.4f);
	glVertex3f(0.01f,-1.0f,-0.4f);
	glVertex3f(0.01f,1.0f,-0.4f);
	glVertex3f(-0.01f,1.0f,-0.4f);
	glEnd();

};

void DiskRender::setFocus(int q, bool b) {
	m_iQuadrant = q;
	m_bFocus = b;
};

Vector * DiskRender::getTextureCoords(int q) {
	return m_pPlane[q].GetTextureCoords();
}
	
	
