#include "Plane.h"

Plane::Plane(){};

Plane::Plane(Vector p, float s) {
	m_vPosition = p;
	m_fScale = s;
	m_vTexCoords= new Vector[4];
};
Vector Plane::GetPosition() {
	return m_vPosition;
};
void Plane::SetPosition(Vector p){
	m_vPosition = p;
};

float Plane::GetScale() {
	return m_fScale;
};
void Plane::SetScale(float s) {
	m_fScale = s;
};

Vector * Plane::GetTextureCoords() {
	return m_vTexCoords;
};

void Plane::SetTextureCoords(Vector * tex) {
	for(int i = 0; i<4;i++) {
		m_vTexCoords[i] = tex[i];
	}
};

/* Render a plane whith its demo
I know it is a bit redundant to bind and release the program and textures all the time
but the renderer does not need all that much performence, and the structure becomes a litle better
*/
void Plane::render(Environment & m_envEnvironment, Program * m_shaderProgram){
	
	float threshold = 0.0;
	bool cutaway = m_envEnvironment.GetCutAwayEnabled();
	
	if (cutaway) {
		threshold = m_envEnvironment.GetCutAwayThreshold();
	}
	float * CA = new float[2];
	CA[0] = (float)cutaway;
	CA[1] = threshold;
	//end Birkeland




	m_shaderProgram->bind();
	glUniform1iARB(m_shaderProgram->GetUniformLocation("samDataVolume"), 0);
	glUniform1iARB(m_shaderProgram->GetUniformLocation("samBackgroundSelectionColorTransferFunction"), 1);
	
	glUniform2fvARB(m_shaderProgram->GetUniformLocation("cutaway"),1,CA);
	


	//bind the volume texture and the color transferfunction
	glActiveTexture(GL_TEXTURE0);
	m_envEnvironment.GetDataVolume().bind();
	glActiveTexture(GL_TEXTURE1);
	m_envEnvironment.GetBackgroundSelectionColorTransferFunction().bind();

	glBegin(GL_QUADS);
	glTexCoord3d(m_vTexCoords[0].GetX(),m_vTexCoords[0].GetZ(),m_vTexCoords[0].GetY());
	glVertex3f(-0.5,-0.5,0.0);

	glTexCoord3d(m_vTexCoords[1].GetX(),m_vTexCoords[1].GetZ(),m_vTexCoords[1].GetY());
	glVertex3f(0.5,-0.5,0.0);

	glTexCoord3d(m_vTexCoords[2].GetX(),m_vTexCoords[2].GetZ(),m_vTexCoords[2].GetY());
	glVertex3f(0.5,0.5,0.0);

	glTexCoord3d(m_vTexCoords[3].GetX(),m_vTexCoords[3].GetZ(),m_vTexCoords[3].GetY());
	glVertex3f(-0.5,0.5,0.0);
	
	glEnd();	


	//release the volumetexture and the color transfer function
	glActiveTexture(GL_TEXTURE1);
	m_envEnvironment.GetBackgroundSelectionColorTransferFunction().release();
	glActiveTexture(GL_TEXTURE0);
	m_envEnvironment.GetDataVolume().release();
	m_shaderProgram->release();

	renderDemo();
	
};


void Plane::renderDemo() {
	float cube[2][4][3] = 
		{
			{
				{-0.5,-0.5,-0.5},
				{0.5,-0.5,-0.5},
				{0.5,0.5,-0.5},
				{-0.5,0.5,-0.5}
			},
			{
				
				{-0.5,-0.5,0.5},
				{0.5,-0.5,0.5},
				{0.5,0.5,0.5},
				{-0.5,0.5,0.5}
			}
		};	
	/*seems very hard coded but it has to be so the litle demo will end 
	up where it is suposed to be
	*/
	glTranslatef(0.35f,-0.35f,-0.3f);
	glScalef(m_fScale/10,m_fScale/10,m_fScale/10);
	
	glRotatef(-10,1,0,0);
	glRotatef(10,0,1,0);
	
	glBegin(GL_QUADS);
	glColor3f(1.0,0,0);
	//front
	glVertex3fv(cube[0][0]);
	glVertex3fv(cube[0][1]);
	glVertex3fv(cube[0][2]);
	glVertex3fv(cube[0][3]);
	
	//left
	glColor3f(0,1,0);
	glVertex3fv(cube[0][1]);
	glVertex3fv(cube[1][1]);
	glVertex3fv(cube[1][2]);
	glVertex3fv(cube[0][2]);

	//top
	glColor3f(0,0,1);
	glVertex3fv(cube[0][2]);
	glVertex3fv(cube[1][2]);
	glVertex3fv(cube[1][3]);
	glVertex3fv(cube[0][3]);
	glEnd();

	//Creating a plane slightlightly larger than the texturecoordinates
	Vector v1 = m_vTexCoords[0] - m_vTexCoords[1];
	Vector v2 = m_vTexCoords[0] - m_vTexCoords[3];

	v1*=0.5;
	v2*=0.5;
	
	Vector plane[4];
	plane[0] = m_vTexCoords[0]+v1+v2;
	plane[1] = m_vTexCoords[1]-v1+v2;
	plane[2] = m_vTexCoords[2]-v1-v2;
	plane[3] = m_vTexCoords[3]+v1-v2;
	
	//texture coordinates goes from 0 to 1, translate it -0.5 so that it will be sentered
	glBegin(GL_QUADS);
	glColor3f(0.5,0.5,0.5);
	glVertex3f(plane[0].GetX()-0.5f,plane[0].GetY()-0.5f,plane[0].GetZ()-0.5f);
	glVertex3f(plane[1].GetX()-0.5f,plane[1].GetY()-0.5f,plane[1].GetZ()-0.5f);
	glVertex3f(plane[2].GetX()-0.5f,plane[2].GetY()-0.5f,plane[2].GetZ()-0.5f);
	glVertex3f(plane[3].GetX()-0.5f,plane[3].GetY()-0.5f,plane[3].GetZ()-0.5f);
	glEnd();
};
